"""Custom exceptions for the lti_toolbox app."""


class LTIException(Exception):
    """Custom LTI exception for proper handling of LTI specific errors."""


class LTIRequestNotVerifiedException(LTIException):
    """
    Custom LTI exception thrown when someone try to access
    LTI parameters before verifying the request.
    """

    def __init__(self):
        message = "You must verify the LTI request with verify() before accessing its parameters"
        super().__init__(message)


class LaunchParamException(Exception):
    """Custom Exception related to LTI launch param processing."""


class InvalidLaunchParamException(LaunchParamException):
    """Custom Exception thrown when an invalid parameter is found in a LTI launch request."""

    def __init__(self, param):
        message = f"{param:s} is not a valid launch param"
        super().__init__(message)


class MissingLaunchParamException(LaunchParamException):
    """
    Custom Exception thrown when a required LTI parameter is missing in
    an LTI launch request.
    """

    def __init__(self, param):
        message = f"missing launch param : {param:s}"
        super().__init__(message)



"""Declare the models related to lti provider."""

import secrets
import string

from django.db import models
from django.utils.translation import gettext_lazy as _


class LTIConsumer(models.Model):
    """
    Model representing a LTI Consumer.
    """

    slug = models.SlugField(
        verbose_name=_("consumer site identifier"),
        primary_key=True,
        help_text=_("identifier for the consumer site"),
    )

    title = models.CharField(
        max_length=255,
        verbose_name=_("Title"),
        help_text=_("human readable title, to describe the LTI consumer"),
        blank=False,
    )

    url = models.URLField(
        max_length=1024,
        verbose_name=_("Consumer site URL"),
        help_text=_("URL of the LTI consumer website"),
        blank=True,
    )

    class Meta:
        """Options for the ``LTIConsumer`` model."""

        db_table = "lti_consumer"
        verbose_name = _("LTI consumer")
        verbose_name_plural = _("LTI consumers")

    def __str__(self):
        """Get the string representation of an instance."""
        return self.title


class LTIPassport(models.Model):
    """
    Model representing an LTI passport for LTI consumers to interact with the django application.

    A LTI consumer can have multiple passports.

    An LTI passport stores credentials that can be used by an LTI consumer to interact with
    the django application acting as an LTI provider.
    """

    consumer = models.ForeignKey(LTIConsumer, on_delete=models.CASCADE)

    title = models.CharField(
        max_length=255,
        verbose_name=_("Title"),
        help_text=_(
            "human readable title, to describe this LTI passport (i.e. : who will use it ?)"
        ),
        blank=False,
    )

    oauth_consumer_key = models.CharField(
        max_length=255,
        verbose_name=_("oauth consumer key"),
        unique=True,
        help_text=_(
            "oauth consumer key to authenticate an LTI consumer on the LTI provider"
        ),
        editable=False,
    )
    shared_secret = models.CharField(
        max_length=255,
        verbose_name=_("shared secret"),
        help_text=_("LTI Shared secret"),
        editable=False,
    )
    is_enabled = models.BooleanField(
        verbose_name=_("is enabled"),
        help_text=_("whether the passport is enabled"),
        default=True,
    )

    class Meta:
        """Options for the ``LTIPassport`` model."""

        db_table = "lti_passport"
        verbose_name = _("LTI passport")
        verbose_name_plural = _("LTI passports")

    def __str__(self):
        """Get the string representation of an instance."""
        return self.title

    # pylint: disable=arguments-differ
    def save(self, *args, **kwargs):
        """Generate the oauth consumer key and shared secret randomly upon creation.

        Parameters
        ----------
        Args:
            args (list) Passed onto parent's `save` method
            dict (kwargs) : Passed onto parent's `save` method

        """
        self.full_clean()
        if not self.oauth_consumer_key:
            self.oauth_consumer_key = self.generate_consumer_key()
        if not self.shared_secret:
            self.shared_secret = self.generate_shared_secret()
        super().save(*args, **kwargs)

    @staticmethod
    def generate_consumer_key() -> str:
        """Generate a random consumer key."""
        oauth_consumer_key_chars = string.ascii_uppercase + string.digits
        oauth_consumer_key_size = secrets.choice(range(20, 30))
        return "".join(
            secrets.choice(oauth_consumer_key_chars)
            for _ in range(oauth_consumer_key_size)
        )

    @staticmethod
    def generate_shared_secret() -> str:
        """Generate a random consumer key."""
        secret_chars = string.ascii_letters + string.digits + "!#$%&*+-=?@^_"
        secret_size = secret_size = secrets.choice(range(40, 60))
        return "".join(secrets.choice(secret_chars) for _ in range(secret_size))



# Generated by Django 3.1.5 on 2021-01-06 03:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("lti_toolbox", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="lticonsumer",
            name="url",
            field=models.URLField(
                blank=True,
                help_text="URL of the LTI consumer website",
                max_length=1024,
                verbose_name="Consumer site URL",
            ),
        ),
    ]



"""Views of the lti_toolbox django application."""
from abc import ABC, abstractmethod

from django.contrib.auth import authenticate, login
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from lti_toolbox.exceptions import LTIException

from .lti import LTI


@method_decorator(csrf_exempt, name="dispatch")
class BaseLTIView(ABC, View):
    """
    Abstract view to help building LTI authenticated views.
    It is convenient if you want to handle dynamic LTI launch URLs.
    """

    def post(self, request, *args, **kwargs) -> HttpResponse:  # pylint: disable=W0613
        """Handler for POST requests."""
        lti_request = LTI(request)
        try:
            lti_request.verify()
            return self._do_on_success(lti_request)
        except LTIException as error:
            return self._do_on_failure(request, error)

    def _do_on_success(self, lti_request: LTI) -> HttpResponse:
        """
        Handler for successfully verified LTI launch requests.
        You have to define your own handler according to your project needs.
        It can be used to call an authentication backend, to redirect the user...etc
        """
        raise NotImplementedError()

    def _do_on_failure(  # pylint: disable=R0201
        self, request: HttpRequest, error: LTIException  # pylint: disable=W0613
    ) -> HttpResponse:
        """
        Default handler for invalid LTI launch requests.
        You are encouraged to define your own handler according to your project needs.
        """
        return HttpResponseForbidden("Invalid LTI launch request")


@method_decorator(csrf_exempt, name="dispatch")
class BaseLTIAuthView(ABC, View):
    """
    Abstract view to help building LTI authenticated views.
    It is convenient if you want to handle dynamic LTI launch URLs.
    """

    def post(self, request, *args, **kwargs) -> HttpResponse:  # pylint: disable=W0613
        """Handler for POST requests."""
        lti_request = LTI(request)
        try:
            lti_request.verify()
            user = authenticate(request, lti_request=lti_request)
            if user is not None:
                login(request, user)
                return self._do_on_login(lti_request)
            return self._do_on_authentication_failure(lti_request)
        except LTIException as error:
            return self._do_on_verification_failure(request, error)

    @abstractmethod
    def _do_on_login(self, lti_request: LTI) -> HttpResponse:
        """Process the request when the user is logged in via LTI"""
        raise NotImplementedError()

    def _do_on_authentication_failure(  # pylint: disable=R0201
        self, lti_request: LTI  # pylint: disable=W0613
    ) -> HttpResponse:
        """
        Default handler for failed authentication.
        You are encouraged to define your own handler according to your project needs.
        """
        return HttpResponseForbidden()

    def _do_on_verification_failure(  # pylint: disable=R0201
        self, request: HttpRequest, error: LTIException  # pylint: disable=W0613
    ) -> HttpResponse:
        """
        Default handler for invalid LTI launch requests.
        You are encouraged to define your own handler according to your project needs.
        """
        return HttpResponseForbidden("Invalid LTI launch request")



"""lti_toolbox application."""
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class LtiProviderAppConfig(AppConfig):
    """Configuration class for the lti_toolbox app."""

    verbose_name = _("LTI Toolbox")
    name = "lti_toolbox"



"""
This module is an implementation of logic needed for checking
OAuth 1.0 signed LTI launch requests.


It is based on the work of django-lti-provider-auth
(https://github.com/wachjose88/django-lti-provider-auth)

Here is the original licence :

Copyright (c) 2018 Josef Wachtler

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""
import logging
import time

from django.core.cache import InvalidCacheBackendError
from django.core.cache import cache as default_cache
from django.core.cache import caches
from oauthlib.oauth1 import RequestValidator

from .models import LTIPassport

logger = logging.getLogger(__name__)


class LTIRequestValidator(RequestValidator):
    """
    This validator implements the RequestValidator from the oauthlib, but only with methods
    required for a LTI launch request.
    """

    LTI_REPLAY_PROTECTION_CACHE = "lti_replay"

    @property
    def enforce_ssl(self):
        """
        Returns:
            bool: True if SSL is mandatory, False otherwise
        """
        return False

    @property
    def nonce_length(self):
        """
        Returns:
            tuple: A tuple containing the min and max length of a nonce
        """
        return 5, 50

    @property
    def dummy_client(self):
        """Dummy client used when an invalid client key is supplied.

        Returns:
            string: The dummy client key string.
        """
        return "dummy_client_key_123456"

    def get_client_secret(self, client_key, request):
        """Retrieves the client secret associated with the client key.

        If an unknown client_key is given, it returns a valid dummy secret in
        order to avoid timing attacks.

        This method must allow the use of a dummy client_key value.
        Fetching the secret using the dummy key must take the same amount of
        time as fetching a secret for a valid client::

        Args:
            client_key: The client/consumer key.
            request: The calling request.
        """

        try:
            passport = LTIPassport.objects.get(
                oauth_consumer_key=client_key, is_enabled=True
            )
            return passport.shared_secret
        except LTIPassport.DoesNotExist:
            return "dummy_client_sec_123456"

    def validate_client_key(self, client_key, request):
        """Validates that supplied client key is a registered and valid client.

        Args:
            client_key: The client/consumer key.
            request: The calling request

        Returns:
            bool: True if the client key is registered and valid
        """
        return LTIPassport.objects.filter(
            oauth_consumer_key=client_key, is_enabled=True
        ).exists()

    # pylint: disable=too-many-arguments
    def validate_timestamp_and_nonce(
        self,
        client_key: str,
        timestamp: str,
        nonce: str,
        request,
        request_token=None,
        access_token=None,
    ):
        """Validates that the nonce has not been used before

        Per `Section 3.3`_ of the OAuth 1.0 spec.

        "A nonce is a random string, uniquely generated by the client to allow
        the server to verify that a request has never been made before and
        helps prevent replay attacks when requests are made over a non-secure
        channel.  The nonce value MUST be unique across all requests with the
        same timestamp, client credentials, and token combinations."

        .. _`Section 3.3`: https://tools.ietf.org/html/rfc5849#section-3.3

        One of the first validation checks that will be made is for the validity
        of the nonce and timestamp, which are associated with a client key and
        possibly a token. If invalid then immediately fail the request
        by returning False. If the nonce/timestamp pair has been used before and
        you may just have detected a replay attack. Therefore it is an essential
        part of OAuth security that you not allow nonce/timestamp reuse.
        Note that this validation check is done before checking the validity of
        the client and token.::

        Args:
            client_key: The client/consumer key.
            timestamp: The ``oauth_timestamp`` request parameter.
            nonce: The ``oauth_nonce`` request parameter.
            request: The calling request
            request_token: Request token string, if any.
            access_token: Access token string, if any.

        Returns:
            bool: True if the timestamp and nonce has not been used before
        """

        cache_timeout = 3600
        # Disallow usage of timestamp older than cache_timeout
        request_timestamp = int(timestamp)
        if request_timestamp < int(time.time()) - cache_timeout:
            logger.debug(
                "Timestamp is too old (ts = %s, consumer_key = %s, nonce = %s)",
                timestamp,
                client_key,
                nonce,
            )
            return False

        try:
            cache = caches[self.LTI_REPLAY_PROTECTION_CACHE]
        except InvalidCacheBackendError:
            logger.debug(
                "Unable to find cache %s, fallback to default cache",
                self.LTI_REPLAY_PROTECTION_CACHE,
            )
            cache = default_cache

        key = f"LTI_TS_NONCE:{client_key:s}:{timestamp:s}:{nonce:s}"

        if not cache.add(key, "1", cache_timeout):
            logger.warning(
                "Replayed timestamp/nonce detected (ts = %s, consumer_key = %s, nonce = %s)",
                timestamp,
                client_key,
                nonce,
            )
            return False

        logger.debug(
            "Timestamp and nonce valid (ts = %s, consumer_key = %s, nonce = %s)",
            timestamp,
            client_key,
            nonce,
        )
        return True



"""Factories for the ``lti_toolbox``."""
import factory
from factory.django import DjangoModelFactory

from . import models


class LTIConsumerFactory(DjangoModelFactory):
    """Factory to create LTI consumer."""

    class Meta:
        model = models.LTIConsumer
        django_get_or_create = ("slug",)

    slug = factory.Sequence(lambda n: f"consumer{n}")
    title = factory.Sequence(lambda n: f"Consumer {n}")
    url = factory.Sequence(lambda n: f"https://www.example.com/consumer-{n}/")


class LTIPassportFactory(DjangoModelFactory):
    """Factory to create LTI passport."""

    class Meta:
        model = models.LTIPassport
        django_get_or_create = (
            "title",
            "consumer",
        )

    title = factory.Sequence(lambda n: f"passport {n}")

    consumer = factory.SubFactory(LTIConsumerFactory)



"""lti_toolbox is a django application to handle LTI Provider related
operations, like launch request verification and user authentication."""

# pylint: disable=invalid-name
default_app_config = "lti_toolbox.apps.LtiProviderAppConfig"



"""
Utilities to represent and validate LTI Launch parameters

This is based on lti library (https://github.com/pylti/lti).
"""
from collections.abc import MutableMapping
from typing import List, Union
from urllib.parse import urlencode

from .exceptions import InvalidLaunchParamException, MissingLaunchParamException

DEFAULT_LTI_VERSION = "LTI-1.0"

LAUNCH_PARAMS_REQUIRED = ["lti_message_type", "lti_version", "resource_link_id"]

LAUNCH_PARAMS_RECOMMENDED = [
    "context_id",
    "context_label",
    "context_title",
    "context_type",
    "launch_presentation_css_url",
    "launch_presentation_document_target",
    "launch_presentation_height",
    "launch_presentation_locale",
    "launch_presentation_return_url",
    "launch_presentation_width",
    "lis_person_contact_email_primary",
    "lis_person_name_family",
    "lis_person_name_full",
    "lis_person_name_given",
    "resource_link_description",
    "resource_link_title",
    "roles",
    "role_scope_mentor",
    "tool_consumer_info_product_family_code",
    "tool_consumer_info_version",
    "tool_consumer_instance_contact_email",
    "tool_consumer_instance_description",
    "tool_consumer_instance_guid",
    "tool_consumer_instance_name",
    "tool_consumer_instance_url",
    "user_id",
    "user_image",
]

LAUNCH_PARAMS_LIS = [
    "lis_course_offering_sourcedid",
    "lis_course_section_sourcedid",
    "lis_outcome_service_url",
    "lis_person_sourcedid",
    "lis_result_sourcedid",
]

LAUNCH_PARAMS_RETURN_URL = [
    "lti_errorlog",
    "lti_errormsg",
    "lti_log",
    "lti_msg",
]

LAUNCH_PARAMS_OAUTH = [
    "oauth_callback",
    "oauth_consumer_key",
    "oauth_nonce",
    "oauth_signature",
    "oauth_signature_method",
    "oauth_timestamp",
    "oauth_token",
    "oauth_version",
]

LAUNCH_PARAMS_IS_LIST = [
    "accept_media_types",
    "accept_presentation_document_targets",
    "context_type",
    "role_scope_mentor",
    "roles",
]

LAUNCH_PARAMS_CANVAS = ["selection_directive", "text"]

CONTENT_PARAMS_REQUEST = [
    "accept_copy_advice",
    "accept_media_types",
    "accept_multiple",
    "accept_presentation_document_targets",
    "accept_unsigned",
    "auto_create",
    "can_confirm",
    "content_item_return_url",
    "data",
    "title",
]

CONTENT_PARAMS_RESPONSE = [
    "content_items",
    "lti_errorlog",
    "lti_errormsg",
    "lti_log",
    "lti_msg",
]

REGISTRATION_PARAMS = [
    "reg_key",
    "reg_password",
    "tc_profile_url",
]

LAUNCH_PARAMS = (
    CONTENT_PARAMS_REQUEST
    + CONTENT_PARAMS_RESPONSE
    + LAUNCH_PARAMS_CANVAS
    + LAUNCH_PARAMS_LIS
    + LAUNCH_PARAMS_OAUTH
    + LAUNCH_PARAMS_RECOMMENDED
    + LAUNCH_PARAMS_REQUIRED
    + LAUNCH_PARAMS_RETURN_URL
    + REGISTRATION_PARAMS
)


class LaunchParams(MutableMapping):
    """
    Represents the params for a LTI launch request. Provides dict-like
    behavior through the use of the MutableMapping ABC mixin.  Strictly
    enforces that params are valid LTI params.
    """

    def __init__(self, *args, **kwargs):

        self._params = dict()
        self.update(*args, **kwargs)

        # now verify we only got valid launch params
        for k in self.keys():
            if not self.valid_param(k):
                raise InvalidLaunchParamException(k)

        for param in LAUNCH_PARAMS_REQUIRED:
            if param not in self:
                raise MissingLaunchParamException(param)

    def _param_value(self, param: str) -> Union[str, List]:
        """Get the value of a LTI parameter.

        Args:
            param: LTI parameter name

        Returns:
            The value of the LTI parameter, as a str or a List, depending on the parameter.
        """
        if param in LAUNCH_PARAMS_IS_LIST:
            return [x.strip() for x in self._params[param].split(",")]
        return self._params[param]

    @staticmethod
    def valid_param(param: str) -> bool:
        """Checks if a LTI parameter is valid or not.

        Args:
            param: LTI parameter name

        Returns:
            bool: True if the parameter is valid, False otherwise.

        """
        if param.startswith("custom_") or param.startswith("ext_"):
            return True
        return param in LAUNCH_PARAMS

    def __len__(self):
        return len(self._params)

    def __getitem__(self, item):
        if not self.valid_param(item):
            raise KeyError("{} is not a valid launch param".format(item))
        try:
            return self._param_value(item)
        except KeyError:
            # catch and raise new KeyError in the proper context
            raise KeyError(item)

    def __setitem__(self, key, value):
        if not self.valid_param(key):
            raise InvalidLaunchParamException(key)
        if key in LAUNCH_PARAMS_IS_LIST:
            if isinstance(value, list):
                value = ",".join([x.strip() for x in value])
        self._params[key] = value

    def __delitem__(self, key):
        if key in self._params:
            del self._params[key]

    def __iter__(self):
        return iter(self._params)

    @property
    def urlencoded(self) -> str:
        """Get the URL encoded representation of the LTI parameter list.

        Returns:
            str: URL encoded LTI parameters
        """
        params = dict(self)
        # stringify any list values
        for key, value in params.items():
            if isinstance(value, list):
                params[key] = ",".join(value)
        return urlencode(params)



"""LTI module that supports LTI 1.0.
"""
import re
from typing import Any, Optional

from oauthlib.oauth1 import SignatureOnlyEndpoint

from .exceptions import (
    LaunchParamException,
    LTIException,
    LTIRequestNotVerifiedException,
)
from .launch_params import LaunchParams
from .models import LTIConsumer, LTIPassport
from .validator import LTIRequestValidator


class LTI:
    """The LTI object abstracts an LTI launch request.

    It provides properties and methods to inspect the launch request.
    """

    def __init__(self, request):
        """Initialize the LTI system.

        Args:
            request (HttpRequest) The request that holds the LTI parameters
        """

        self.request = request
        self._valid = None
        self._params = {}
        self._verified = False

    def verify(self) -> bool:
        """Verify the LTI request.

        Returns:
            bool: True if the LTI launch request is valid.

        Raises:
            LTIException: Raised if request validation fails
        """

        try:
            launch_params = LaunchParams(self.request.POST)
        except LaunchParamException as error:
            raise LTIException(
                "Exception while processing launch parameters : {}".format(error)
            )

        validator = LTIRequestValidator()
        oauth_endpoint = SignatureOnlyEndpoint(validator)

        self._valid, _ = oauth_endpoint.validate_request(
            uri=self.request.build_absolute_uri(),
            http_method=self.request.method,
            body=launch_params.urlencoded,
            headers=self.request.headers,
        )

        if self._valid is not True:
            raise LTIException("LTI verification failed")

        self._params = launch_params
        self._verified = True

        return self._valid

    @property
    def is_valid(self) -> bool:
        """Check if the LTI request is verified and valid

        Returns:
            True if the request is verified and valid
        """
        return self._verified and self._valid

    def get_param(self, name: str, default: Any = None):
        """Retrieve a LTI parameter value given its name.

        Args:
            name: Name of the LTI parameter
            default: Default value if the parameter is not found

        Returns:
            The value of the LTI parameter if it exists, or the default value otherwise.

        """
        if not self._verified:
            raise LTIRequestNotVerifiedException()
        return self._params.get(name, default)

    def get_consumer(self) -> LTIConsumer:
        """Retrieve the LTI consumer that initiated the launch request."""
        consumer_key = self.get_param("oauth_consumer_key")
        passport = LTIPassport.objects.get(
            oauth_consumer_key=consumer_key, is_enabled=True
        )
        return passport.consumer

    def get_course_info(self) -> dict:
        """Retrieve course info in the LTI request.

        Returns:
            dict: A dictionnary containing course informations
                school_name: the school name
                course_name: the course name
                course_section: the course section

        """
        if self.is_edx_format:
            groups = re.match(r"^course-v[0-9]:(.*)", self.get_param("context_id"))
            if groups is not None:
                part = groups.group(1).split("+")
                length = len(part)
                return {
                    "school_name": part[0] if length >= 1 else None,
                    "course_name": part[1] if length >= 2 else None,
                    "course_run": part[2] if length >= 3 else None,
                }

        return {
            "school_name": self.get_param("tool_consumer_instance_name", None),
            "course_name": self.get_param("context_title"),
            "course_run": None,
        }

    @property
    def resource_link_title(self) -> Optional[str]:
        """Return the resource link id as default for its title."""
        return self.get_param("resource_link_title", self.get_param("resource_link_id"))

    @property
    def context_title(self) -> Optional[str]:
        """Return the context id as default for its title."""
        return self.get_param("context_title", self.get_param("context_id"))

    @property
    def roles(self):
        """LTI roles of the authenticated user.

        Returns:
            List[str]: normalized LTI roles from the session
        """
        roles = self.get_param("roles", [])
        return list(map(str.lower, roles))

    @property
    def is_edx_format(self):
        """Check if the LTI request comes from Open edX.

        Returns:
            boolean: True if the LTI request is sent by Open edX

        """
        return re.search(r"^course-v[0-9]:.*$", self.get_param("context_id"))
